import requests
import re
from bs4 import BeautifulSoup
import json
from concurrent.futures import ThreadPoolExecutor
from deprecated import deprecated

def get_video_id_from_url(url):
    pattern = r'(?:v=|\/)([0-9A-Za-z_-]{11})'
    match = re.search(pattern, url)
    return match.group(1) if match else None

# deprecated: 放弃纯正则匹配, 容易受到转义字符干扰, 改为使用 json 解析更加准确
# 改良版在get_youtube_title_by_url
@deprecated(reason="Use get_youtube_title_by_url instead")
def _get_youtube_title_by_url(url):
    import regex  # 第三方库，性能更强
    TITLE_PATTERN = regex.compile(r'"playerOverlayVideoDetailsRenderer":\{"title":\{"simpleText":"(.*?)"')
    try:
        response = requests.get(url, timeout=(30, 60))  # 设置连接和读取超时
        match = TITLE_PATTERN.search(response.text)  # 直接使用编译后的对象
        return match.group(1) if match else None
    except Exception as e:
        print(f"Error fetching {url}: {e}")
        return None


def get_youtube_title_by_url(url):
    try:
        response = requests.get(url, timeout=(30, 60))  # 设置连接和读取超时
        soup = BeautifulSoup(response.text, 'html.parser')
        scripts = soup.find_all("script", {'nonce': True})
        for script in scripts:
            if 'playerOverlayVideoDetailsRenderer' in script.text:
                json_str = script.string.strip()
                json_data = json.loads(json_str[json_str.find('{'):json_str.rfind('}')+1])
                title = json_data["playerOverlays"]\
                    ["playerOverlayRenderer"]\
                    ["videoDetails"]\
                    ["playerOverlayVideoDetailsRenderer"]\
                    ["title"]["simpleText"]
                return title
        return None
    except Exception as e:
        print(f"Error fetching {url}: {e}")
        return None


def get_youtube_title_by_url_batch(urls, max_workers=None):
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        return list(executor.map(get_youtube_title_by_url, urls))


def get_youtube_author_by_url(url):
    try:
        response = requests.get(url, timeout=(30, 60))  # 设置连接和读取超时
        soup = BeautifulSoup(response.text, 'html.parser')
        scripts = soup.find_all("script", {'nonce': True})

        for script in scripts:
            if 'playerOverlayVideoDetailsRenderer' in script.text:
                json_str = script.string.strip()
                json_data = json.loads(json_str[json_str.find('{'):json_str.rfind('}')+1])
                author = json_data["playerOverlays"]\
                    ["playerOverlayRenderer"]\
                    ["videoDetails"]\
                    ["playerOverlayVideoDetailsRenderer"]\
                    ["subtitle"]["runs"][0]["text"]
                return author
        return None
    except Exception as e:
        print(f"Error fetching {url}: {e}")
        return None


def get_youtube_author_by_url_batch(urls, max_workers=None):
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        return list(executor.map(get_youtube_author_by_url, urls))

def get_youtube_title_by_oembed(id_or_url):
    api_prefix = "https://www.youtube.com/oembed?url="
    try:
        if "youtube.com/watch?v=" in id_or_url:
            video_id = get_video_id_from_url(id_or_url)
            url = api_prefix + f"https://www.youtube.com/watch?v={video_id}&format=json"
        elif len(id_or_url) == 11:
            url = api_prefix + f"https://www.youtube.com/watch?v={id_or_url}&format=json"
        else:
            raise ValueError("Invalid YouTube ID or URL format")
        response = requests.get(url, timeout=(30, 60))  # 设置连接和读取超时
        data = response.json()
        return data.get("title")
    except Exception as e:
        print(f"Error fetching oEmbed for {url}: {e}")
        return None
    

def get_youtube_title_by_oembed_batch(ids_or_urls, max_workers=None):
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        return list(executor.map(get_youtube_title_by_oembed, ids_or_urls))


def get_youtube_info_by_oembed(id_or_url):
    api_prefix = "https://www.youtube.com/oembed?url="
    try:
        if "youtube.com/watch?v=" in id_or_url:
            video_id = get_video_id_from_url(id_or_url)
            url = api_prefix + f"https://www.youtube.com/watch?v={video_id}&format=json"
        elif len(id_or_url) == 11:
            url = api_prefix + f"https://www.youtube.com/watch?v={id_or_url}&format=json"
        else:
            raise ValueError("Invalid YouTube ID or URL format")
        response = requests.get(url, timeout=(30, 60))  # 设置连接和读取超时
        data = response.json()
        return data
    except Exception as e:
        print(f"Error fetching oEmbed for {url}: {e}")
        return None
    
def get_youtube_info_by_oembed_batch(ids_or_urls, max_workers=None):
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        return list(executor.map(get_youtube_info_by_oembed, ids_or_urls))


# -*- coding: utf-8 -*-
import importlib

# 想延迟加载的子模块列表
_LAZY_SUBMODULES = ['ai', 'cv', 'data', 'dataset', 'loanlib']

# 需要在顶层暴露的函数映射：顶层名字 -> (子模块名, 子模块内函数名)
_EXPOSED_FUNCS = {
    'fast_loadenv_then_append_path': ('loanlib', 'fast_loadenv_then_append_path'),
    'o_d': ('misc.timer', 'o_d'),
    'O_D': ('misc.timer', 'O_D'),
    'timer': ('misc.timer', 'timer'),
    # 如果以后想暴露更多函数，继续添加
    # 'other_func': ('loanlib', 'other_func'),
}

# 缓存已加载的子模块
_loaded_submodules = {}

def __getattr__(name):
    # 先检查是否是暴露的函数
    if name in _EXPOSED_FUNCS:
        mod_name, func_name = _EXPOSED_FUNCS[name]
        if mod_name not in _loaded_submodules:
            _loaded_submodules[mod_name] = importlib.import_module(f'.{mod_name}', __name__)
        func = getattr(_loaded_submodules[mod_name], func_name)
        # 缓存到globals，避免重复导入
        globals()[name] = func
        return func

    # 再检查是否是延迟加载的子模块
    if name in _LAZY_SUBMODULES:
        if name not in _loaded_submodules:
            _loaded_submodules[name] = importlib.import_module(f'.{name}', __name__)
        module = _loaded_submodules[name]
        globals()[name] = module
        return module

    raise AttributeError(f"module {__name__} has no attribute {name}")

def whoami():
    """Who am I?"""
    return "I am Elinor, a Python package for data processing and analysis."



import os
import sys
import ffmpeg
from concurrent.futures import ThreadPoolExecutor

import logging
import argparse
from timeit import default_timer as timer
import csv


def timeit(func):
    def wrapper(*args, **kwargs):
        start = timer()
        result = func(*args, **kwargs)
        end = timer()
        logging.info(f"Function {func.__name__} took {end - start:.2f} seconds")
        return result
    return wrapper


def convert_mp4_to_mp3(input_video, output_audio, cmd='ffmpeg'):
    logging.info(f"{input_video} converting...")
    if os.path.exists(output_audio):
        logging.info(f"{output_audio} already exists.")
        return 0
    
    try:
        (
            ffmpeg
            .input(input_video)
            .output(
                output_audio,
                **{
                    'q:a': 2,          # 2表示较好质量，范围0-9
                    'c:a': 'libmp3lame',
                    # 'preset': 'fast',  # 编码预设
                    'threads': '2',    # 分配2个线程
                    'vn': None         # 禁用视频流处理
                }
            )
            # .overwrite_output()  # 如果输出文件已存在，则覆盖
            .run(quiet=True, cmd=cmd)           # 禁用控制台输出
        )
        logging.info(f"{output_audio} created successfully.")
        return 0
    except Exception as e:
        logging.error(f"Error converting {input_video}: {e}")
        return 1


@timeit
def convert_videos_in_folder(input_folder, output_folder, cmd='ffmpeg'):
    files = []
    for filename in os.listdir(input_folder):
        if filename.endswith(".mp4"):
            input_video = os.path.join(input_folder, filename)
            output_audio = os.path.join(output_folder, f"{os.path.splitext(filename)[0]}.mp3")
            files.append((input_video, output_audio, cmd))
    
    max_workers = os.cpu_count() * 2 if os.cpu_count() else 4
    
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = [executor.submit(convert_mp4_to_mp3, *file) for file in files]
        for future in futures:
            result = future.result()


def ffprobe_get_duration(input_video, cmd='ffprobe'):
    try:
        probe = ffmpeg.probe(input_video, cmd=cmd)
        duration = float(probe['format']['duration'])
        return duration
    except ffmpeg.Error as e:
        logging.error(f"Error probing {input_video}: {e}")
        return None


def cos_to_csv(input_path, output_path, sep="\x1F"):
    with open(input_path, 'r', encoding='utf-8') as f:
        raw = f.readlines()
    csv_part = raw[2: -3]  # remove the first two and last three lines
    
    tmp = []
    for i, line in enumerate(csv_part):
        parts = line.rsplit("|", 5)
        if len(parts) == 6:
            tmp.append(parts)
        else:
            raise ValueError(f"Line {i+2} does not have exactly 6 parts: {line}")

    with open(output_path, "w", newline='') as f:
        writer = csv.writer(f, delimiter=sep)
        writer.writerow(["KEY", "TYPE", "LAST MODIFIED", "ETAG", "SIZE", "RESTORESTATUS"])
        writer.writerows(tmp)


import time
import functools
import asyncio
from typing import Callable, Any

from datetime import datetime, timezone, timedelta
from functools import wraps

def memoize_first_call_with_reset(func):
    cached_result = None
    has_been_called = False
    
    @wraps(func)
    def wrapper(*args, **kwargs):
        nonlocal cached_result, has_been_called
        
        # 检查是否需要重置缓存
        reset = kwargs.pop('_reset', False)
        if reset:
            cached_result = None
            has_been_called = False
        
        if not has_been_called:
            cached_result = func(*args, **kwargs)
            has_been_called = True
        
        return cached_result
    
    # 添加重置方法
    def reset():
        nonlocal cached_result, has_been_called
        cached_result = None
        has_been_called = False
        
    wrapper.reset = reset
    return wrapper

@memoize_first_call_with_reset
def O_D (td: int = 0, tz: timezone = None):
    if tz is not None:
        return datetime.now(tz)
    if td != 0:
        tz = timezone(timedelta(hours=td))
        return datetime.now(tz)
    now = datetime.now()
    return now.astimezone()


def o_d(td:int=0, tz:timezone=None):
    if tz is not None:
        return datetime.now(tz)
    if td !=0:
        tz = timezone(datetime.timedelta(hours=td))
        return datetime.now(tz)
    now = datetime.now()
    return now.astimezone()


def timer(func: Callable) -> Callable:
    """
    极简版计时装饰器，支持同步和异步函数
    """
    
    @functools.wraps(func)
    def sync_wrapper(*args, **kwargs) -> Any:
        start = time.perf_counter()
        try:
            return func(*args, **kwargs)
        finally:
            end = time.perf_counter()
            print(f"函数 {func.__name__} 执行耗时: {end - start:.4f} 秒")
    
    @functools.wraps(func)
    async def async_wrapper(*args, **kwargs) -> Any:
        start = time.perf_counter()
        try:
            return await func(*args, **kwargs)
        finally:
            end = time.perf_counter()
            print(f"函数 {func.__name__} 执行耗时: {end - start:.4f} 秒")
    
    return async_wrapper if asyncio.iscoroutinefunction(func) else sync_wrapper


if __name__ == "__main__":
    @timer
    def sync_function(x: int) -> int:
        time.sleep(x)
        return x * 2

    @timer
    async def async_function(x: int) -> int:
        await asyncio.sleep(x)
        return x * 3

    # 测试同步函数
    print(sync_function(2))

    # 测试异步函数
    print(asyncio.run(async_function(2)))
    



from io import BytesIO
from PIL import Image
import base64

def pil_to_b64(img):
    buf = BytesIO()
    img.save(buf, format="PNG")
    # default codec is utf-8
    return base64.b64encode(buf.getvalue()).decode()

def b64_to_pil(b64_str):
    return Image.open(BytesIO(base64.b64decode(b64_str)))

if __name__ == "__main__":
    # Example usage
    img = Image.open("../data/elinor.jpeg")
    img_str = pil_to_b64(img)
    print("Encoded string:", img_str)
    
    decoded_img = b64_to_pil(img_str)
    decoded_img.show()


# TODO:
import os
import pandas as pd
from typing import List, Generator, Optional
from elinor import count_files_by_end
from PIL import Image
# from dataclasses import dataclass
from dataclasses import dataclass
from string import Template
from tqdm import tqdm
from types import SimpleNamespace
import numpy as np

"""
TemplateFeatureDir = Template("${root_dir}/features/fullFrame-210x260px/${split}")
TemplateAnnotation = Template("${root_dir}/annotations/manual/PHOENIX-2014-T.${split}.corpus.csv")
TemplateFrame = Template("images${frame}.png")
"""
T = SimpleNamespace(
    FeatureDir = Template("${root_dir}/features/fullFrame-210x260px/${split}"),
    Annotation = Template("${root_dir}/annotations/manual/PHOENIX-2014-T.${split}.corpus.csv"),
    Frame = Template("images${frame}.png")
)


@dataclass
class BaseDatasetPhoenix2014TItem:
    frames: List[Image.Image]
    paths: List[str]
    num: int
    orth: str
    translation: str
    poses: Optional[np.ndarray]
    dirname: Optional[str]


class BaseDatasetPhoenix2014T(object):
    """Phoenix-2014-T数据集基础类

    数据集结构大致如下：
    PHOENIX-2014-T-release-v3/
    └── PHOENIX-2014-T       <------- ROOT DIR
        ├── annotations
        │   └── manual
        ├── evaluation
        │   ├── sign-recognition
        │   └── sign-translation
        ├── features
        │   └── fullFrame-210x260px
        │       ├── dev
        │       ├── test
        │       └── train
        └── models
            └── languagemodels
    """
    def __init__(
            self, 
            root_dir="./PHOENIX-2014-T",
            split="dev" # train, test, dev
        ):
        self.root_dir=root_dir
        self.features_dir = T.FeatureDir.substitute(root_dir=self.root_dir,split=split)
        self.annotations_path = T.Annotation.substitute(root_dir=self.root_dir, split=split)
        self.metadata = self.generate_metadata_from_annotations()


    def __len__(self):
        return len(self.metadata)


    def __getitem__(self, index):
        metadata = self.metadata.iloc[index]
        frames =  self.get_frames_by_dirname(metadata.dirname)
        paths = self.get_paths_by_dirname(metadata.dirname)
        num = self.metadata.iloc[index].num

        return BaseDatasetPhoenix2014TItem(
            frames=frames,
            paths=paths,
            num=num,
            orth=metadata.orth,
            translation=metadata.translation,
            dirname=metadata.dirname,
            poses=None,
        )


    def generate_metadata_from_annotations(self):
        metadata = pd.read_csv(self.annotations_path, sep="|").rename(columns={"name": "dirname"})
        
        frame_nums = []
        for dirname in tqdm(metadata.dirname):
            # Check if the directory exists
            dir_path = os.path.join(self.features_dir, dirname)
            frame_num = len([f for f in os.listdir(dir_path) if f.endswith('.png')])
            frame_nums.append(frame_num)
        
        metadata["num"] = frame_nums
        return metadata 
    

    def get_frames_by_dirname(self, dirname) -> Generator:
        video_path = os.path.join(self.features_dir, dirname)
        frame_num = self.metadata[self.metadata["dirname"] == dirname].num.item()
        frames = [""] * frame_num
        for i in range(frame_num):
            frame_path = os.path.join(video_path, T.Frame.substitute(frame=f"{i+1:04d}"))
            frames[i] = Image.open(frame_path)
        return frames


    def get_paths_by_dirname(self, dirname) -> Generator:
        video_path = os.path.join(self.features_dir, dirname)
        frame_num = self.metadata[self.metadata["dirname"] == dirname].num.item()
        paths = [""] * frame_num
        for i in range(frame_num):
            frame_path = os.path.join(video_path, T.Frame.substitute(frame=f"{i+1:04d}"))
            paths[i] = frame_path
        return paths

# TODO: 需要实现一个数据集的子类，来实现对姿态数据的处理
class AdvancedDatasetPhoenix2014T(BaseDatasetPhoenix2014T):
    def __init__(self, root_dir="./PHOENIX-2014-T", split="dev"):
        super().__init__(root_dir=root_dir, split=split)
        self.poses_dir = os.path.join(self.root_dir, "features", "pose", split)
        self.metadata = self.generate_metadata_from_annotations()
        pass


import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from tqdm import tqdm
from functools import partial


def _count_files_in_dir(ends, args):
    root, files = args
    return sum(1 for file in files if file.lower().endswith(ends))


def count_files_by_end(directory, ends=".png", workers=None):
    """
    统计目录下所有文件的数量，支持多线程并行处理。
    该函数会遍历指定目录及其子目录，统计所有文件的数量。
    Args:
        directory: 要扫描的目录路径
        workers: 并行工作线程数(默认使用CPU核心数)
    """

    if workers is None:
        from multiprocessing import cpu_count
        workers = max(cpu_count() - 1, 1)  # 保留一个核心用于其他任务

    # 收集所有目录和文件
    walk_results = [(root, files) for root, _, files in os.walk(directory)]

    func = partial(_count_files_in_dir, ends)

    total = 0
    with ThreadPoolExecutor(max_workers=workers) as executor:
        futures = [executor.submit(func, args) for args in walk_results]
        for f in tqdm(as_completed(futures), total=len(futures), desc="Scanning Dirs", unit="dir", mininterval=1):
            total += f.result()

    return total


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("directory", help="Directory to scan")
    parser.add_argument("-j", "--workers", type=int, default=None,
                        help="Number of worker threads")
    parser.add_argument("-e", "--ends", type=str, default=".png",
                        help="File extension to count (default: .png)")
    args = parser.parse_args()

    files_count = count_files_by_end(args.directory, ends=args.ends, workers=args.workers)
    print(f"Total files: {files_count}")


from rapidfuzz import fuzz, process

def match_best_str(
    target: str,
    choices: list[str],
    score_cutoff: int = 70,
) -> tuple[str, float]:
    output = process.extractOne(target, choices, scorer=fuzz.ratio, score_cutoff=score_cutoff)
    if output is None:
        return "", 0 # Return empty string and score 0 if no match found
    best_match, best_score, _ = output
    return best_match, best_score



from setuptools import find_packages, setup

setup(
    name="elinor",
    version="0.0.21",
    author="Dashvvood",
    author_email="mathismottis@gmail.com",
    description="some util functions",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "tqdm",
        "dotenv",
        "deprecated",
    ],
    extra_require={
        "dev": [
            "pybase64",
            "pillow",
            "pandas",
            "pyyaml",
            "bs4",
            "requests",
            "regex",
            "rapidfuzz",
            "ffmpeg-python"
        ]
    }
)


from .count_files import count_files_by_end
from .string_matcher import match_best_str
from .convert import convert_mp4_to_mp3, convert_videos_in_folder
from .youtube import (
    get_video_id_from_url,
    get_youtube_title_by_url_batch,
    get_youtube_author_by_url_batch,
    get_youtube_title_by_oembed_batch,
    get_youtube_info_by_oembed_batch
)


from collections import defaultdict, OrderedDict, namedtuple, Counter
import uuid

def get_input_output_per_layer(model, model_input):
    features = OrderedDict()
    handles = OrderedDict()
    HookData = namedtuple("HookData", ["type", "input", "output"])
    
    def hook(name):
        def hook_fn(module, input, output):
            features[name] = HookData(type(module), input, output)
        return hook_fn
    
    for name, module in model.named_modules():
        handles[name] = module.register_forward_hook(hook(name))
    
    with torch.no_grad():
        model(**model_input)
        
    for name, h in handles.items():
        h.remove()
        
    return features

def get_mapping_f2p(features, parameters):
    """
    Get a mapping from feature to parameter.
    """
    def get_main_key(key):
        return key.rsplit(".", 1)[0]  # 删除最后一个 '.' 后的部分（即 .weight, .bias）

    # 建立映射关系
    mapping = defaultdict(list)
    for p_key in parameters.keys():
        main_key = get_main_key(p_key)
        for f_key in features.keys():
            if f_key.startswith(main_key):
                mapping[f_key].append(p_key)
        
    return mapping

def summary(model, model_input):
    features = get_input_output_per_layer(model, model_input)
    parameters = {}
    for k, v in model.named_parameters():
        parameters[k] = v
    mapping = get_mapping_f2p(features, parameters)

    uuid.uuid4()
    fp = open(f"summary_{uuid.uuid4().hex}.txt", "w")

    fp.write("Model: " + type(model).__name__ + "\n")
    fp.write("Input: " + str(model_input.keys()) + "\n")
    fp.write("=" * 80 + "\n")

    for f_key, io_data in features.items():
        input_shape = io_data.input[0].shape if len(io_data.input) > 0 else None
        output_shape = io_data.output[0].shape if len(io_data.output) > 0 else None
        module_type = io_data.type.__name__
        fp.write(f"{f_key}: --- {module_type}\n")
        fp.write(" " * 4 + f">>> {input_shape} --> {output_shape}\n")
        
        for p_key in mapping[f_key]:
            fp.write(" "*4 + f"{p_key}: {parameters[p_key].shape}\n")
            
        fp.write("-" * 80 + "\n")
    fp.write("=" * 80 + "\n")
    fp.close()
    
    return features, parameters, mapping


if __name__ == '__main__':
    from diffusers import UNet2DModel
    import torch
    
    model = UNet2DModel()
    model_input = {
        "sample": torch.randn(1, 3, 64, 64),
        "timestep": 1,
    }
    
    summary(model, model_input)
        


import os
import sys
from typing import List
from dotenv import load_dotenv, find_dotenv, dotenv_values
from types import SimpleNamespace
from deprecated import deprecated

def fast_loadenv_then_append_path(keys:List[str]=["PROJECT_ROOT"], verbose=False) -> SimpleNamespace:

    env_path = find_dotenv(usecwd=True)
    load_res = load_dotenv(dotenv_path=env_path)

    if keys is not None and len(keys) > 0:
        for key in keys:
            value = os.getenv(key)
            if value and os.path.exists(value) and os.path.isdir(value):
                sys.path.append(value)
                if verbose:
                    print(f"appending {value} to sys.path")

    # env_values = dotenv_values(dotenv_path=env_path)
    # if append:
    #     for _, value in env_values.items():
    #         if os.path.exists(value) and os.path.isdir(value):
    #             sys.path.append(value)
    env_values = dotenv_values(dotenv_path=env_path)
    ns = SimpleNamespace(**env_values)
    return ns

